# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2016-03-16 19:26:49.214985
from __future__ import unicode_literals
import datetime

version = "1.1.1"
githash = "fedd7c52ba80e2246f83b8d1752baa3a153a33e2"


major = 1
minor = 1
bugfix = 1

release = True
timestamp = datetime.datetime(2016, 3, 16, 19, 26, 49, 214985)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
